"""This module extends a DeviceGraph to enable SST Simulation output."""

import os
import orjson
from .Device import *
from .DeviceGraph import *


class SSTGraph(DeviceGraph):
    """
    SSTGraph is an extension to DeviceGraph that lets you build or write JSON.

    SSTGraph will flatten the graph when building or writing JSON. You probably
    want to do these last since they modify the DeviceGraph itself
    """

    def __init__(self, graph: DeviceGraph) -> None:
        """Point at the DeviceGraph variables."""
        self.expanding = None
        self.attr = graph.attr
        self.devices = graph.devices
        self.names = graph.names
        self.links = graph.links
        self.ports = graph.ports

    def build(self, nranks: int = 1) -> dict:
        """
        Build the SST graph.

        Return a dictionary of component names to SST component objects.
        If you have an extremely large graph, it is recommended that you use
        ahp_graph to do the graph partitioning instead of letting SST do it. You
        can do this by using the Device.set_partition() function and then
        setting nranks in this function to the total number of ranks used
        """
        # only import sst when we are going to build the graph
        import sst

        # If this is serial or sst is doing the partitioning,
        # generate the entire graph
        if nranks == 1:
            self.flatten()
            self.verify_links()
            return self.__build_model(False)

        # Find the partition information and build the model for this rank
        else:
            sst.setProgramOption("partitioner", "sst.self")
            rank = sst.getMyMPIRank()
            # If this is an extra rank, don't do anything
            if rank >= nranks:
                return dict()

            self.check_partition()
            self.flatten(rank=rank)
            # need to verify before we follow links because pruning
            # may make the overall graph invalid
            # (it will be valid for the current rank)
            self.verify_links()
            self.follow_links(rank, True)
            return self.__build_model(True)

    def write_json(self, filename: str, nranks: int = 1,
                   program_options: dict = None) -> None:
        """
        Generate the JSON and write it to the specified filename.

        All output will be stored in a folder called output
        The program_options dictionary provides a way to pass SST
        program options, such as timebase and stopAtCycle.

        If you have an extremely large graph, it is recommended that you use
        ahp_graph to do the graph partitioning instead of letting SST do it. You
        can do this by using the Device.set_partition() function and then
        setting nranks in this function to the total number of ranks used
        """
        if not os.path.exists('output'):
            os.makedirs('output')

        self.flatten()
        self.verify_links()

        # If this is serial or sst is doing the partitioning,
        # just write the whole thing
        if nranks == 1:
            self.__write_model(f"output/{filename}", nranks,
                               self.devices, self.links, program_options)

        # Write a JSON file for each rank
        else:
            (base, ext) = os.path.splitext(f"output/{filename}")
            self.check_partition()
            partition = self.__partition_graph(nranks)
            for rank in range(nranks):
                self.__write_model(base + str(rank) + ext, nranks,
                                   partition[rank][0], partition[rank][1],
                                   program_options)

    def __partition_graph(self, nranks: int) -> list:
        """
        Store all the devices and links for a given rank.

        Return a list of pairs of the form (Device-set, link-dict),
        where each list entry corresponds to a particular processor rank.
        It is faster to partition the graph once rather than search it
        O(p) times.
        """
        partition = [(set(), dict()) for p in range(nranks)]

        for p0, p1 in self.links:
            link = frozenset({p0, p1})
            t = self.links[link]
            d0 = p0.device
            d1 = p1.device
            while d0.subOwner is not None:
                d0 = d0.subOwner
            while d1.subOwner is not None:
                d1 = d1.subOwner
            r0 = d0.partition[0]
            r1 = d1.partition[0]

            partition[r0][0].add(d0)
            partition[r0][0].add(d1)
            partition[r0][1][link] = t
            if r0 != r1:
                partition[r1][0].add(d0)
                partition[r1][0].add(d1)
                partition[r1][1][link] = t
        return partition

    @staticmethod
    def __encode(attr: dict,
                 stringify: bool = False) -> dict:
        """
        Convert attributes into SST Params.

        SST only supports primitive types (bool, int, float) and lists of those
        types as parameters; everything else we will convert via JSON.
        If the attribute contains a __to_json__ method, then we will call it.
        Ignore bad conversions.
        """

        def supported_f(x) -> bool:
            """Return whether the type is supported by SST."""
            return isinstance(x, (bool, float, int, str))

        params = dict()
        for (key, val) in attr.items():
            native = supported_f(val)
            if not native and isinstance(val, list):
                native = all(map(supported_f, val))

            if native:
                params[key] = val if not stringify else str(val)
            elif hasattr(val, "__to_json__"):
                params[key] = val.__to_json__()
            else:
                try:
                    # serialize the value to json bytes,
                    # then deserialize into a python dict
                    params[key] = orjson.loads(
                        orjson.dumps(val, option=orjson.OPT_INDENT_2)
                    )
                except Exception:
                    pass

        return params

    def __build_model(self, self_partition: bool) -> dict:
        """Generate the model for the SST program."""
        # only import sst when we are going to build the graph inside of sst
        import sst
        n2c = dict()

        # Set up global parameters.
        global_params = self.__encode(self.attr)
        for (key, val) in global_params.items():
            sst.addGlobalParam(key, key, val)

        def recurseSubcomponents(dev: Device, comp: 'sst.Component') -> None:
            """Add subcomponents to the Device."""
            for (d1, n1, s1) in dev.subs:
                if d1.library is None:
                    raise RuntimeError(f"No SST library: {d1.name}")
                if s1 is None:
                    c1 = comp.setSubComponent(n1, d1.library)
                else:
                    c1 = comp.setSubComponent(n1, d1.library, s1)
                c1.addParams(self.__encode(
                    {'type': d1.type, 'model': d1.model} | d1.attr))
                n2c[d1.name] = c1
                for key in global_params:
                    c1.addGlobalParamSet(key)
                if len(d1.subs) > 0:
                    recurseSubcomponents(d1, c1)

        # First, we instantiate all of the components with
        # their attributes. Ignore Devices that have no library defined
        for d0 in self.devices:
            if d0.subOwner is None and d0.library is not None:
                c0 = sst.Component(d0.name, d0.library)
                c0.addParams(self.__encode(
                    {'type': d0.type, 'model': d0.model} | d0.attr))
                # Set the component partition if we are self-partitioning
                if self_partition:
                    thread = (0 if d0.partition[1] is None
                              else d0.partition[1])
                    c0.setRank(d0.partition[0], thread)
                n2c[d0.name] = c0
                for key in global_params:
                    c0.addGlobalParamSet(key)
                recurseSubcomponents(d0, c0)

        # Second, link the component ports using graph links
        for p0, p1 in self.links:
            t = self.links[frozenset({p0, p1})]
            if p0.device.library is not None and p1.device.library is not None:
                c0 = n2c[p0.device.name]
                c1 = n2c[p1.device.name]
                s0 = p0.get_name()
                s1 = p1.get_name()
                if str(p0) < str(p1):
                    link = sst.Link(f'{p0}__{t}__{p1}')
                else:
                    link = sst.Link(f'{p1}__{t}__{p0}')
                latency = t if t != '0s' else '1ps'
                link.connect((c0, s0, latency), (c1, s1, latency))

        # Return a map of component names to components.
        return n2c

    def __write_model(self, filename: str, nranks: int, devices: set,
                      links: dict,
                      program_options: dict = None) -> None:
        """Write this DeviceGraph out as JSON."""
        model = dict()

        # Write the program options to the model.
        if program_options is None:
            model["program_options"] = dict()
        else:
            model["program_options"] = dict(program_options)

        # If running in parallel, then set up the SST SELF partitioner.
        if nranks > 1:
            model["program_options"]["partitioner"] = "sst.self"

        # Set up global parameters.
        global_params = self.__encode(self.attr, True)
        model["global_params"] = dict()
        for (key, val) in global_params.items():
            model["global_params"][key] = dict({key: val})
        global_set = list(global_params.keys())

        def recurseSubcomponents(dev: Device) -> list:
            """Add subcomponents to the Device."""
            subcomponents = list()
            for (d1, n1, s1) in dev.subs:
                if d1.library is None:
                    raise RuntimeError(f"No library: {d1.name}")

                item = {
                        "slot_name": n1,
                        "type": d1.library,
                        "slot_number": s1,
                        "params": self.__encode(
                            {'type': d1.type, 'model': d1.model} | d1.attr,
                            True),
                        "params_global_sets": global_set,
                    }
                if len(d1.subs) > 0:
                    item["subcomponents"] = recurseSubcomponents(d1)
                subcomponents.append(item)
            return subcomponents

        # Define all the components. We define the name, type, parameters,
        # and global parameters. Ignore Devices that have no library defined
        components = list()
        for d0 in devices:
            if d0.subOwner is None and d0.library is not None:
                component = {
                    "name": d0.name,
                    "type": d0.library,
                    "params": self.__encode(
                        {'type': d0.type, 'model': d0.model} | d0.attr,
                        True),
                    "params_global_sets": global_set,
                }
                if nranks > 1:
                    component["partition"] = {
                        "rank": d0.partition[0],
                        "thread": (0 if d0.partition[1] is None
                                   else d0.partition[1]),
                    }

                subcomponents = recurseSubcomponents(d0)
                if len(subcomponents) > 0:
                    component["subcomponents"] = subcomponents
                components.append(component)

        model["components"] = components

        # Now define the links between components.
        linksJSON = list()
        for p0, p1 in links:
            if p0.device.library is not None and p1.device.library is not None:
                t = links[frozenset({p0, p1})]
                latency = t if t != '0s' else '1ps'
                if str(p0) < str(p1):
                    name = f'{p0}__{t}__{p1}'
                else:
                    name = f'{p1}__{t}__{p0}'
                linksJSON.append(
                    {
                        "name": name,
                        "left": {
                            "component": p0.device.name,
                            "port": p0.get_name(),
                            "latency": latency,
                        },
                        "right": {
                            "component": p1.device.name,
                            "port": p1.get_name(),
                            "latency": latency,
                        },
                    }
                )

        model["links"] = linksJSON

        with open(filename, "wb") as jfile:
            jfile.write(orjson.dumps(model, option=orjson.OPT_INDENT_2))



"""
This module represents a Device.

There are decorators for annotating which library the device can be found
in along with port labels.

The most important feature is that a Device can be an assembly, meaning it is
constructed of other Devices. This combined with the DeviceGraph allows for
hierarchical representations of a graph.
"""

import collections


class PortInfo(dict):
    """
    PortInfo is a dictionary describing what ports a Device can have.

    PortInfo contains information such as port type, limit on the number of
    connections, whether the port is required, and formatting info.
    PortInfo is a class variable which defines the available ports for all
    instances of a specific Device. DevicePort represents an actual instance of
    a port on a Device.
    """

    def add(self, name: str, ptype: str = None, limit: int = 1,
            required: bool = True, format: str = '.#') -> None:
        """Add a port definition to the dictionary."""
        self[name] = (limit, ptype, required, format)


class DevicePort:
    """
    A DevicePort represents an instance of a port on a Device.

    DevicePort contains a Device reference, a port name, and an
    optional port number.
    It can also reference the other DevicePort that it is linked to.
    """

    def __init__(self, device: 'Device', name: str,
                 number: int = None) -> None:
        """Initialize the device, name, port number."""
        self.device = device
        self.name = name
        self.number = number
        self.link = None

    def get_name(self) -> str:
        """Return a string representation of the port name and number."""
        if self.number is None:
            return self.name
        else:
            sf = self.device.portinfo[self.name][3].split('#')
            return f"{self.name}{sf[0]}{self.number}{sf[1]}"

    def __repr__(self) -> str:
        """Return a string representation of this DevicePort."""
        return f"{self.device.name}.{self.get_name()}"


class Device:
    """
    Device is the base class for a node in a DeviceGraph.

    A Device may be represented by an model or may be an assembly
    of other Devices. If an assembly, then the Device must define
    an expand() method that returns a DeviceGraph that implements the Device.

    Devices may contain submodules which allow for parameterization of
    functionality similar to a lambda function (ex. SST subcomponents). This
    is done by creating a Device to represent the submodule and then adding
    it into another Device.

    The variables library and portinfo are class variables that can be set on
    the definition of a new Device class. The variable attr is a dictionary of
    attributes that can be used as both a class variable and an instance
    variable. Attributes that are common among all instances of a Device can
    be set in the definition of that Device and instance attributes can be
    added to the class attributes on specific instances of the Device.
    """

    library = None
    portinfo = PortInfo()
    attr = dict()

    def __init__(self, name: str, model: str = None,
                 attr: dict = None) -> None:
        """
        Initialize the Device.

        Initialize with the unique name, model, and optional
        dictionary of attributes which are used as model parameters.
        """
        self.name = name
        if self.attr:
            if attr is not None:
                self.attr = {**self.attr, **attr}
        elif attr is not None:
            self.attr = attr
        self.ports = collections.defaultdict(dict)
        self.subs = set()
        self.subOwner = None
        self.partition = None
        self.type = self.__class__.__name__
        self.model = model
        if self.library is None and not hasattr(self, "expand"):
            raise RuntimeError(f"Assemblies must define expand(): {self.type}")

    def set_partition(self, rank: int, thread: int = None) -> None:
        """Assign a rank and optional thread to this device."""
        self.partition = (rank, thread)

    def add_submodule(self, device: 'Device', slotName: str,
                      slotIndex: int = None) -> None:
        """
        Add a submodule to this Device.

        Both the submodule and this Device must have libraries.
        Note that all submodules must be added to the Device before
        the Device is added to a DeviceGraph.
        """
        if self.library is None or device.library is None:
            raise RuntimeError(f"Submodule and parent must have libraries:"
                               f" {device.name}, {self.name}")
        device.subOwner = self
        self.subs.add((device, slotName, slotIndex))

    def __getattr__(self, port: str) -> 'DevicePort':
        """
        Enable ports to be treated as variables.

        As a convenience, we allow ports to be named as an attribute on the
        class (e.g., Device.Input instead of Device.port("Input").
        If the port is not defined, then we thrown an exception.
        """
        info = self.portinfo.get(port)
        if info is None:
            raise RuntimeError(f"Unknown port in {self.name}: {port}")

        elif info[0] == 1:
            return self.port(port)
        else:
            return lambda x: self.port(port, x)

    def port(self, port: str, number: int = None) -> 'DevicePort':
        """
        Return a DevicePort object representing the port on this Device.

        If a Single port, then make sure we do not have a port number.
        If the port has not already been defined, then add it.

        If a limited or multi port, determine the port number.  If
        number is None, then create a new port at the end of the current
        list.  Make sure we do not create too many connections if limited.
        Finally, if the port has not already been defined, then create it.
        """
        info = self.portinfo.get(port)
        if info is None:
            raise RuntimeError(f"Unknown port in {self.name}: {port}")

        elif info[0] == 1:
            # Single Port, don't use port numbers
            if number is not None:
                print(f"WARNING! Single port ({port}) being provided a port"
                      f" number ({number})")
            if port not in self.ports:
                self.ports[port][-1] = DevicePort(self, port)
            return self.ports[port][-1]

        else:
            # Multi Port, use port numbers and check the provided limit
            if number is None:
                number = len(self.ports[port])
            if info[0] is not None:
                if number >= info[0]:
                    raise RuntimeError(f"Too many connections ({number}):"
                                       f" {port} (limit {info[0]})")
            if number not in self.ports[port]:
                self.ports[port][number] = DevicePort(self, port, number)
            return self.ports[port][number]

    def get_category(self) -> str:
        """Return the category for this Device (type, model)."""
        if self.model is not None:
            return f"{self.type}_{self.model}"
        return self.type

    def label_ports(self) -> str:
        """Return the port labels for a graphviz record style node."""
        if len(self.ports) == 0:
            return ''
        label = '{'
        for port in self.ports:
            label += f"<{port}> {port}|"
        return label[:-1] + '}'

    def __repr__(self) -> str:
        """Return a description of the Device."""
        lines = list()
        lines.append(f"Device = {self.type}")
        lines.append(f"\tname = {self.name}")
        if self.model:
            lines.append(f"\tmodel = {self.model}")
        if self.library:
            lines.append(f"\tlibrary = {self.library}")
        else:
            lines.append(f"\tassembly = True")
        if self.partition:
            lines.append(f"\tpartition = {self.partition}")
        if self.subOwner is not None:
            lines.append(f"\tsubmoduleParent = {self.subOwner.name}")

        if self.portinfo:
            lines.append(f"\tPorts:")
            for port in sorted(self.portinfo):
                lines.append(f"\t\t{port} = {self.portinfo[port]}")
        if self.attr:
            lines.append(f"\tAttributes:")
            for key in sorted(self.attr):
                lines.append(f"\t\t{key} = {self.attr[key]}")
        if self.subs:
            lines.append(f"\tSubmodules:")
            for sub in sorted(self.subs, key=lambda x: (x[1], x[2])):
                lines.append(f"\t\t{sub[1]}:{sub[2]} -> {sub[0].name}")

        return "\n".join(lines)



"""Collection of ahp_graph Devices for testing."""

from ahp_graph.Device import *
from ahp_graph.DeviceGraph import *


class LibraryTestDevice(Device):
    """Unit test for Library."""

    library = 'ElementLibrary.Component'

    def __init__(self, name: str = '') -> None:
        """Test Device for Library."""
        super().__init__(f'{self.__class__.__name__}{name}')


class PortTestDevice(Device):
    """Unit test for Device Ports."""

    library = 'ElementLibrary.Component'
    portinfo = PortInfo()
    portinfo.add('default')
    portinfo.add('ptype', 'test')
    portinfo.add('no_limit', limit=None)
    portinfo.add('limit', limit=2)
    portinfo.add('optional', required=False)
    portinfo.add('format', limit=None, format='(#)')
    portinfo.add('exampleSinglePort', 'test', required=False)
    portinfo.add('exampleMultiPort', 'test', None, False)
    portinfo.add('port', 'port', None, False)

    def __init__(self, name: str = '') -> None:
        """Test Device for Device Ports."""
        super().__init__(f'{self.__class__.__name__}{name}')


class LibraryPortTestDevice(Device):
    """Unit test for Device with Library and Ports."""

    library = 'ElementLibrary.Component'
    portinfo = PortInfo()
    portinfo.add('input')
    portinfo.add('output')
    portinfo.add('optional', limit=None, required=False)

    def __init__(self, name: str = '') -> None:
        """Test Device for Library and Ports."""
        super().__init__(f'{self.__class__.__name__}{name}')


class RecursiveAssemblyTestDevice(Device):
    """Unit Test for a recursive assembly. Creates a ring of Devices."""

    portinfo = PortInfo()
    portinfo.add('input')
    portinfo.add('output')
    portinfo.add('optional', limit=None, required=False)

    def __init__(self, levels: int, name: str = '') -> None:
        """Test Device for recursive assembly."""
        super().__init__(f'{self.__class__.__name__}{levels}{name}', str(levels))

    def expand(self, graph: DeviceGraph) -> None:
        """Test for expanding a Device."""
        # If we have reached the specified number of levels, create 'leafs'
        # using LibraryPortTestDevice
        level = 0
        if self.model is not None:
            level = int(self.model)
        if level == 0:
            d0 = LibraryPortTestDevice('0')
            d1 = LibraryPortTestDevice('1')
            graph.link(d0.optional(0), self.optional(0))
            graph.link(d1.optional(0), self.optional(1))

        else:
            d0 = RecursiveAssemblyTestDevice(level-1, '0')
            d1 = RecursiveAssemblyTestDevice(level-1, '1')
            for i in range(2 ** level):
                graph.link(d0.optional(i), self.optional(i*2))
                graph.link(d1.optional(i), self.optional(i*2 + 1))

        graph.link(self.input, d0.input)
        graph.link(d0.output, d1.input)
        graph.link(d1.output, self.output)


class ModelTestDevice(Device):
    """Unit test for Device with model."""

    library = 'ElementLibrary.Component'

    def __init__(self, model: str, name: str = '') -> None:
        """Test Device with model."""
        super().__init__(f'{self.__class__.__name__}{name}', model)


class AttributeTestDevice(Device):
    """Unit test for Device with attributes."""

    library = 'ElementLibrary.Component'
    attr = {'a1': 1, 'a2': 'blue', 'a3': False}

    def __init__(self, attr: dict, name: str = '') -> None:
        """Test Device with attributes."""
        super().__init__(f'{self.__class__.__name__}{name}', attr=attr)



"""Simple example of two Python functions playing pingpong with messages."""

from ahp_graph.DeviceGraph import *
from ahp_graph.SSTGraph import *
from architecture import architecture


class Ping():
    """Example class that sends a ping."""

    def __init__(self, name: str, repeats: int) -> None:
        """Init."""
        self.name = name
        self.max = repeats
        self.repeats = 0
        self.output = None

    def start(self) -> None:
        """Start the pingpong off by sending ping."""
        print(f'{self.name}: Sending Ping')
        self.output(f'{self.name}: Ping')

    def input(self, string: str) -> None:
        """Print input and then send ping if we have more repeats left."""
        print(f'{self.name}: Received {string}')
        self.repeats += 1
        if self.repeats < self.max:
            print(f'{self.name}: Sending Ping')
            self.output(f'{self.name}: Ping')


class Pong():
    """Example class that sends a pong when it recieves input."""

    def __init__(self, name: str) -> None:
        """Init."""
        self.name = name
        self.output = None

    def input(self, string: str) -> None:
        """Print input and then send pong."""
        print(f'{self.name}: Received {string}')
        print(f'{self.name}: Sending Pong')
        self.output(f'{self.name}: Pong')


def buildPython(graph: DeviceGraph) -> None:
    """
    Run the example 'simulation' using the two classes.

    Need to manually build the graph since ahp_graph doesn't support this.
    """
    devs = dict()
    # instantiate all the devices in the graph
    for device in graph.devices:
        if device.type == 'Ping':
            devs[device.name] = eval(f'{device.library.split(".")[1]}('
                                     f'"{device.name}", {args.repeats})')
        elif device.type == 'Pong':
            devs[device.name] = eval(f'{device.library.split(".")[1]}('
                                     f'"{device.name}")')
        else:
            print('ERROR, should only have Ping or Pong Devices')
            exit()

    # if we only have one pingpong, manually connect them
    if args.num <= 1:
        pi = devs['PingPong0.Ping']
        po = devs['PingPong0.Pong']
        pi.output = lambda x: po.input(x)
        po.output = lambda x: pi.input(x)
    else:
        def wrapper(dev):
            """Need to wrap the lambda function to protect the device scope."""
            return lambda x: dev.input(x)

        # connect the pingpongs together as specified
        # flip the direction of the one connection since it wraps around
        for p0, p1 in graph.links:
            name0 = p0.device.name
            name1 = p1.device.name
            n0 = int(''.join(filter(str.isdigit, name0)))
            n1 = int(''.join(filter(str.isdigit, name1)))
            t0 = name0.split('.')[1]
            t1 = name1.split('.')[1]
            # by default, we are going to have the first device as output and
            # the second as the input, switch name0 and name1 if that is wrong
            if n0 == n1:
                if t0 == 'Ping' and t1 == 'Pong':
                    pass
                elif t1 == 'Ping' and t0 == 'Pong':
                    (name0, name1) = (name1, name0)
                else:
                    print(f'Numbers equal {n0}, but types'
                          ' do not match {t0}, {t1}')
            else:
                # different PingPong assemblies
                # handle the wrap around case connecting to PingPong0.Ping
                if name0 == 'PingPong0.Ping' and n1 == args.num-1:
                    (name0, name1) = (name1, name0)
                elif name1 == 'PingPong0.Ping' and n0 == args.num-1:
                    pass
                elif n1 > n0:
                    pass
                else:
                    (name0, name1) = (name1, name0)
            devs[name0].output = wrapper(devs[name1])

    # find the first Ping device and start the 'simulation'
    for device in graph.devices:
        if device.type == 'Ping':
            devs[device.name].start()
            break


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description='PingPong')
    parser.add_argument('--num', type=int, default=2,
                        help='how many pingpongs to include')
    parser.add_argument('--repeats', type=int, default=5,
                        help='how many message volleys to run')
    args = parser.parse_args()

    # Construct a DeviceGraph with the specified architecture
    graph = architecture(args.repeats, args.num)

    # generate a graphviz dot file including the hierarchy
    graph.write_dot('pingpong', draw=True, ports=True)

    # flatten the graph and generate a graphviz dot file
    graph.flatten()
    graph.write_dot('pingpongFlat', draw=True, ports=True, hierarchy=False)

    buildPython(graph)



"""Collection of Unit tests for ahp_graph DeviceGraph."""

from ahp_graph.Device import *
from ahp_graph.DeviceGraph import *
from Devices import *


def test_attribute() -> None:
    """Test of DeviceGraph with attributes."""
    attr = {'a1': 1, 'a2': 'blue', 'a3': False}
    graph = DeviceGraph(attr)
    assert graph.attr == attr, 'attr'


def test_addDevice() -> None:
    """Test of adding Devices to a DeviceGraph."""
    graph = DeviceGraph()

    ltd = LibraryTestDevice('ltd')
    sub1 = LibraryTestDevice('sub1')
    sub2 = LibraryTestDevice('sub2')
    sub11 = LibraryTestDevice('sub11')
    sub1.add_submodule(sub11, 'slot')
    ltd.add_submodule(sub1, 'slot', 1)
    ltd.add_submodule(sub2, 'slot', 2)
    graph.add(ltd)

    assert len(graph.devices) == 4, 'num devices'
    assert ltd in graph.devices, 'get device'
    assert sub1 in graph.devices, 'get device'
    assert sub2 in graph.devices, 'get device'
    assert sub11 in graph.devices, 'get device'

    ltd = LibraryTestDevice()
    graph.add(ltd)
    graph.add(ltd)
    assert len(graph.devices) == 5, 'duplicate add'

    sameName = None
    try:
        graph.add(LibraryTestDevice())
        sameName = False
    except RuntimeError:
        sameName = True
    assert sameName, 'same name'


def test_countDevices() -> None:
    """Test of counting Devices in a DeviceGraph."""
    graph = DeviceGraph()

    devs = dict()
    for i in range(10):
        for j in range(3):
            devs[f'mtd{i}.{j}'] = ModelTestDevice(f'model{i}',
                                                  f'mtd{i}.{j}')
            graph.add(devs[f'mtd{i}.{j}'])

    assert len(graph.devices) == 30, 'num devices'
    assert devs['mtd0.0'] in graph.devices, 'get device'
    c = graph.count_devices()
    assert len(c) == 10, 'device count length'
    assert c[devs['mtd0.0'].get_category()] == 3, 'device count length'


def test_checkPartition() -> None:
    """Test of checking partition info in a DeviceGraph."""
    graph = DeviceGraph()

    devs = dict()
    for i in range(10):
        for j in range(3):
            devs[f'mtd{i}.{j}'] = ModelTestDevice(f'model{i}',
                                                  f'mtd{i}.{j}')
            devs[f'mtd{i}.{j}'].set_partition(i, j)
            graph.add(devs[f'mtd{i}.{j}'])

    partitionIncluded = None
    try:
        graph.check_partition()
        partitionIncluded = True
    except RuntimeError:
        partitionIncluded = False
    assert partitionIncluded, 'correct partitioning'

    graph.add(LibraryTestDevice())
    partitionIncluded = None
    try:
        graph.check_partition()
        partitionIncluded = False
    except RuntimeError:
        partitionIncluded = True
    assert partitionIncluded, 'missing partitioning'


def test_link() -> None:
    """Test of linking Devices in a DeviceGraph."""
    graph = DeviceGraph()

    ptd0 = PortTestDevice('0')
    ptd1 = PortTestDevice('1')
    ltd = LibraryTestDevice()
    lptd = LibraryPortTestDevice()
    ltd.add_submodule(lptd, 'slot')

    graph.link(lptd.input, ptd0.optional)
    assert len(graph.devices) == 3, 'linking add submodule parent'
    assert ltd in graph.devices, 'submodule parent included'
    assert graph.links[frozenset({lptd.input, ptd0.optional})] == '0s', 'default latency'
    linkAgain = None
    try:
        graph.link(ptd0.optional, lptd.input)
        linkAgain = False
    except RuntimeError:
        linkAgain = True
    assert linkAgain, 'link again with ports reversed'
    callablePort = None
    try:
        graph.link(ptd0.limit, ptd1.no_limit)
        callablePort = False
    except RuntimeError:
        callablePort = True
    assert callablePort, 'forgot to include port num on multi port'
    typeMismatch = None
    try:
        graph.link(ptd0.default, ptd1.ptype)
        typeMismatch = False
    except RuntimeError:
        typeMismatch = True
    assert typeMismatch, 'port type mismatch'
    changeSinglePortLink = None
    try:
        graph.link(ptd0.optional, ptd1.optional)
        changeSinglePortLink = False
    except RuntimeError:
        changeSinglePortLink = True
    assert changeSinglePortLink, 'linking from a single port again'
    graph.link(ptd0.limit(0), ptd1.limit(0), '123ns')
    assert graph.links[frozenset({ptd0.limit(0), ptd1.limit(0)})] == '123ns', 'latency'


def test_verifyLinks() -> None:
    """Test of verifying links in a DeviceGraph."""
    graph = DeviceGraph()

    ptd0 = PortTestDevice('0')
    ptd1 = PortTestDevice('1')
    ltd = LibraryTestDevice()
    lptd = LibraryPortTestDevice()
    ltd.add_submodule(lptd, 'slot')

    graph.link(lptd.input, ptd0.optional)
    graph.link(lptd.output, ptd1.optional)
    graph.link(ptd0.default, ptd1.default)
    graph.link(ptd0.ptype, ptd1.ptype)
    graph.link(ptd0.no_limit(0), ptd1.no_limit(0))
    graph.link(ptd0.limit(0), ptd1.limit(0))
    verified = None
    try:
        graph.verify_links()
        verified = False
    except RuntimeError:
        verified = True
    assert verified, 'not all required ports connected'

    graph.link(ptd0.format(0), ptd1.format(0))
    verified = None
    try:
        graph.verify_links()
        verified = True
    except RuntimeError:
        verified = False
    assert verified, 'verified all required ports have at least 1 connection'


def test_followLinks() -> None:
    """Test of following links by rank in a DeviceGraph."""
    graph = DeviceGraph()
    levels = 2

    ratd0 = RecursiveAssemblyTestDevice(levels, '0')
    ratd1 = RecursiveAssemblyTestDevice(levels, '1')
    lptd0 = LibraryPortTestDevice('0')
    lptd1 = LibraryPortTestDevice('1')

    # complete the rings
    graph.link(ratd0.input, ratd0.output)
    graph.link(ratd1.input, ratd1.output)

    graph.link(lptd0.input, lptd1.output)
    graph.link(lptd1.input, lptd0.output)
    for i in range(2 ** (levels+1)):
        graph.link(lptd0.optional(None), ratd0.optional(None))
        graph.link(lptd1.optional(None), ratd1.optional(None))

    ratd0.set_partition(0)
    ratd1.set_partition(1)
    lptd0.set_partition(2)
    lptd1.set_partition(2)

    graph.follow_links(0)
    for dev in graph.devices:
        if dev.library is None:
            assert dev.name == f'RecursiveAssemblyTestDevice{levels}1', 'only one assembly left'
        else:
            assert dev.library is not None, 'library set'


def test_flatten() -> None:
    """Test of flattening a DeviceGraph."""
    levels = 6

    def createGraph() -> tuple:
        """Create a graph for testing."""
        graph = DeviceGraph()

        ratd0 = RecursiveAssemblyTestDevice(levels, '0')
        ratd1 = RecursiveAssemblyTestDevice(levels, '1')
        lptd0 = LibraryPortTestDevice('0')
        lptd1 = LibraryPortTestDevice('1')

        # complete the rings
        graph.link(ratd0.input, ratd0.output)
        graph.link(ratd1.input, ratd1.output)

        graph.link(lptd0.input, lptd1.output)
        graph.link(lptd1.input, lptd0.output)
        for i in range(2 ** (levels+1)):
            graph.link(lptd0.optional(None), ratd0.optional(None))
            graph.link(lptd1.optional(None), ratd1.optional(None))

        ratd0.set_partition(0)
        ratd1.set_partition(1)
        lptd0.set_partition(2)
        lptd1.set_partition(2)
        return graph, ratd0

    flat, _ = createGraph()
    flat.flatten()
    assert not any([d.library is None for d in flat.devices]), 'no assemblies left'

    twoLevels, _ = createGraph()
    twoLevels.flatten(2)
    assert [d.library is None for d in twoLevels.devices].count(True) == 8, 'correct number of assemblies left'

    byName, _ = createGraph()
    byName.flatten(name=f'RecursiveAssemblyTestDevice{levels}0')
    rank0, _ = createGraph()
    rank0.flatten(rank=0)
    assert byName.names == rank0.names, 'name and rank devices'

    byNameLinks = set()
    rankLinks = set()
    for link in byName.links:
        for port in link:
            byNameLinks.add(str(port))
    for link in rank0.links:
        for port in link:
            rankLinks.add(str(port))
    assert byNameLinks == rankLinks, 'name and rank links'

    expand, ratd0 = createGraph()
    expand.flatten(expand={ratd0})
    assemblies = [d.library is None for d in expand.devices]
    assert assemblies.count(True) == 3, 'correct number of assemblies left'



"""
Vanadis Processor assembly.

Constructed from Vanadis components along with various L1 Caches.
Processor connects via a memHierarchy Bus.
"""
from ahp_graph.Device import *
from ahp_graph.DeviceGraph import *
import os


class VanadisCPU(Device):
    """VanadisCPU."""

    library = 'vanadis.dbg_VanadisCPU'
    portinfo = PortInfo()
    portinfo.add('os_link', 'os')
    attr = {
        "clock": "1GHz",
        "verbose": 0,
        "physical_fp_registers": 168,
        "physical_int_registers": 180,
        "integer_arith_cycles": 2,
        "integer_arith_units": 2,
        "fp_arith_cycles": 8,
        "fp_arith_units": 2,
        "branch_unit_cycles": 2,
        "print_int_reg": 1,
        "pipeline_trace_file": "",
        "reorder_slots": 64,
        "decodes_per_cycle": 4,
        "issues_per_cycle": 4,
        "retires_per_cycle": 4,
        "pause_when_retire_address": 0
    }


class VanadisMIPSDecoder(Device):
    """VanadisMIPSDecoder."""

    library = 'vanadis.VanadisMIPSDecoder'
    attr = {
        "uop_cache_entries": 1536,
        "predecode_cache_entries": 4
    }


class VanadisMIPSOSHandler(Device):
    """VanadisMIPSOSHandler."""

    library = 'vanadis.VanadisMIPSOSHandler'
    attr = {
        "verbose": 0,
        "brk_zero_memory": "yes"
    }


class VanadisBasicBranchUnit(Device):
    """VanadisBasicBranchUnit."""

    library = 'vanadis.VanadisBasicBranchUnit'
    attr = {
        "branch_entries": 32
    }


class VanadisSequentialLoadStoreQueue(Device):
    """VanadisSequentialLoadStoreQueue."""

    library = 'vanadis.VanadisSequentialLoadStoreQueue'
    attr = {
        "verbose": 0,
        "address_mask": 0xFFFFFFFF,
        "load_store_entries": 32,
        "fault_non_written_loads_after": 0,
        "check_memory_loads": "no",
        "allow_speculated_operations": "no"
    }


class VanadisNodeOS(Device):
    """VanadisNodeOS."""

    library = 'vanadis.VanadisNodeOS'
    portinfo = PortInfo()
    portinfo.add('core', 'os', limit=None, format='#')
    attr = {
        "verbose": 0,
        "heap_start": 512 * 1024 * 1024,
        "heap_end": (2 * 1024 * 1024 * 1024) - 4096,
        "page_size": 4096,
        "heap_verbose": 0,
        "executable": (f"{os.getenv('SST_ELEMENTS_HOME')}"
                       "/src/sst/elements/rdmaNic/tests/app/rdma/msg"),
        "app.argc": 4,
        "app.arg0": "IMB",
        "app.arg1": "PingPong",
        "app.arg2": "-iter",
        "app.arg3": "3",
        "app.env_count": 1,
        "app.env0": f"HOME={os.getenv('HOME')}",
        "app.env1": "VANADIS_THREAD_NUM=0"
    }

    def __init__(self, name: str, cores: int = 1,
                 attr: dict = None) -> None:
        """Initialize with the number of cores per server."""
        super().__init__(name, attr=attr)
        self.attr['cores'] = cores


class memInterface(Device):
    """memInterface."""

    library = 'memHierarchy.standardInterface'
    portinfo = PortInfo()
    portinfo.add('port', 'simpleMem', required=False)

    def __init__(self, name: str, coreId: int = 0,
                 attr: dict = None) -> None:
        """Initialize."""
        super().__init__(name, attr=attr)
        self.attr['coreId'] = coreId


class Cache(Device):
    """Cache."""

    library = 'memHierarchy.Cache'
    portinfo = PortInfo()
    portinfo.add('high_network', 'simpleMem', None, False, '_#')
    portinfo.add('low_network', 'simpleMem', None, False, '_#')
    attr = {
        "replacement_policy": "lru",
        "coherence_protocol": "MESI",
        "cache_line_size": 64,
        "cache_frequency": "1GHz"
    }

    def __init__(self, name: str, model: str, attr: dict = None) -> None:
        """Initialize with a model of which cache level this is (L1, L2)."""
        if model == 'L1':
            parameters = {
                "L1": 1,
                "access_latency_cycles": 2,
                "associativity": 8,
                "cache_size": "32KB",
            }
        elif model == 'L2':
            parameters = {
                "L1": 0,
                "access_latency_cycles": 14,
                "associativity": 16,
                "cache_size": "1MB"
            }
        else:
            return None

        if attr is not None:
            parameters.update(attr)
        super().__init__(name, model, parameters)


class Bus(Device):
    """Bus."""

    library = 'memHierarchy.Bus'
    portinfo = PortInfo()
    portinfo.add('high_network', 'simpleMem', None, False, '_#')
    portinfo.add('low_network', 'simpleMem', None, False, '_#')
    attr = {
        "bus_frequency": "1GHz",
    }


class MemLink(Device):
    """MemLink."""

    library = 'memHierarchy.MemLink'
    portinfo = PortInfo()
    portinfo.add('port', 'simpleMem')


class Processor(Device):
    """Processor assembly made of various Vanadis components and Caches."""

    portinfo = PortInfo()
    portinfo.add('low_network', 'simpleMem', None, False, '_#')

    def __init__(self, name: str, core: int = 0, cores: int = 1,
                 attr: dict = None) -> None:
        """Initialize with the core ID and number of cores in a server."""
        super().__init__(name, attr=attr)
        self.attr['core'] = core
        self.attr['cores'] = cores

    def expand(self, graph: DeviceGraph) -> None:
        """Expand the server into its components."""
        # Reminder to add subcomponents to the main component BEFORE adding
        # the main component to the DeviceGraph. You can also link directly
        # to the subcomponents after they are connected to a parent component
        # Device names created by an assembly will automatically have the
        # assembly name prefixed to the name provided.
        cpu = VanadisCPU("VanadisCPU")

        decoder = VanadisMIPSDecoder("VanadisMIPSDecoder")
        osHandler = VanadisMIPSOSHandler("VanadisMIPSOSHandler")
        branch = VanadisBasicBranchUnit("VanadisBasicBranchUnit")
        decoder.add_submodule(osHandler, "os_handler")
        decoder.add_submodule(branch, "branch_unit")
        cpu.add_submodule(decoder, 'decoder0')

        icache = memInterface("ICache", self.attr['core'])
        cpu.add_submodule(icache, 'mem_interface_inst')

        lsq = VanadisSequentialLoadStoreQueue(
            "VanadisSequentialLoadStoreQueue")
        cpu.add_submodule(lsq, 'lsq')

        dcache = memInterface("DCache", self.attr['core'])
        lsq.add_submodule(dcache, 'memory_interface')

        nodeOS = VanadisNodeOS("VanadisNodeOS", self.attr['cores'])
        nodeOSmem = memInterface("NodeOSMemIF", self.attr['core'])
        nodeOS.add_submodule(nodeOSmem, 'mem_interface')

        nodeOSL1D = Cache("nodeOSL1D", 'L1')

        cpuL1D = Cache("cpuL1D", 'L1')
        cpu_to_L1D = MemLink("cpu_to_L1D")
        L1D_to_L2 = MemLink("L1D_to_L2")
        cpuL1D.add_submodule(cpu_to_L1D, 'cpulink')
        cpuL1D.add_submodule(L1D_to_L2, 'memlink')

        cpuL1I = Cache("cpuL1I", 'L1')
        cpu_to_L1I = MemLink("cpu_to_L1I")
        L1I_to_L2 = MemLink("L1I_to_L2")
        cpuL1I.add_submodule(cpu_to_L1I, 'cpulink')
        cpuL1I.add_submodule(L1I_to_L2, 'memlink')

        bus = Bus("Bus")

        # Linking automatically adds the devices as needed
        graph.link(dcache.port('port'), cpu_to_L1D.port('port'), '1ns')
        graph.link(icache.port('port'), cpu_to_L1I.port('port'), '1ns')
        graph.link(nodeOSmem.port('port'), nodeOSL1D.high_network(0), '1ns')
        graph.link(L1D_to_L2.port('port'), bus.high_network(0), '1ns')
        graph.link(L1I_to_L2.port('port'), bus.high_network(1), '1ns')
        graph.link(nodeOSL1D.low_network(0), bus.high_network(2), '1ns')
        graph.link(cpu.os_link, nodeOS.core(0), '5ns')

        # Our external connection goes through the memHierarchy Bus
        # Generally you don't want to put latency on the links to assembly
        # ports (ex: self.port) and allow whatever uses the assembly to
        # specify latency for the connection (it will get ignored anyway)
        graph.link(bus.low_network(0), self.low_network(0))



"""ahp_graph architecture describing PingPong."""

from ahp_graph.Device import *
from ahp_graph.DeviceGraph import *


class Ping(Device):
    """Ping Device: has a Name and a Model type."""

    library = 'pingpong.Ping'
    portinfo = PortInfo()
    portinfo.add('input', 'String')
    portinfo.add('output', 'String')


class Pong(Device):
    """Pong Device."""

    library = 'pingpong.Pong'
    portinfo = PortInfo()
    portinfo.add('input', 'String')
    portinfo.add('output', 'String')


class pingpong(Device):
    """Assembly of a Ping and Pong device with connections outside."""

    portinfo = PortInfo()
    portinfo.add('input', 'String')
    portinfo.add('output', 'String')

    def expand(self, graph: DeviceGraph) -> None:
        """Expand the overall architecture into its components."""
        # create a Ping Device
        # Device names created by an assembly will automatically have the
        # assembly name prefixed to the name provided.
        ping = Ping('Ping', self.model)
        pong = Pong('Pong')  # create a Pong Device

        # link ping and pong, automatically adds the devices to the graph
        graph.link(ping.output, pong.input, '1s')

        # Generally you don't want to put latency on the links to assembly
        # ports (ex: self.port) and allow whatever uses the assembly to
        # specify latency for the connection (it will get ignored anyway)

        # Connect pingpong's input to ping's input
        graph.link(ping.input, self.input)
        # Connect pingpong's output to pong's output
        graph.link(pong.output, self.output)


def architecture(repeats: int = 10, num: int = 1) -> DeviceGraph:
    """Create pingpong(s) and link them together in a loop."""
    graph = DeviceGraph()
    pingpongs = dict()

    for i in range(num):
        pingpongs[i] = pingpong(f"PingPong{i}", str(repeats))
        pingpongs[i].set_partition(i)

    for i in range(num):
        graph.link(pingpongs[i].output, pingpongs[(i+1) % num].input, '2s')

    return graph



"""
This module implements support for AHP (Attributed Hierarchical Port) graphs.

This class of graph supports attributes on the nodes (Devices), links that are
connected to named ports on the nodes, and nodes that may be represented by a
hierarchical graph (aka an assembly).
All links are bidirectional.
"""

import os
import collections
import pygraphviz
from .Device import *


class DeviceGraph:
    """
    A DeviceGraph is a graph of Devices and their connections to one another.

    The Devices are nodes and the links connect the DevicePorts on the nodes.
    This implements an AHP (Attributed Hierarchical Port) graph.
    """

    def __init__(self, attr: dict = None) -> None:
        """
        Define an empty DeviceGraph.

        The attributes are considered global parameters shared by all instances
        in the graph. They are only supported at the top-level graph,
        not intemediate graphs (e.g., assemblies).
        The dictionary of links uses a frozenset of DevicePorts as the key
        """
        self.expanding = None
        self.attr = attr if attr is not None else dict()
        self.devices = set()
        self.names = set()
        self.links = dict()
        self.ports = set()

    def __repr__(self) -> str:
        """Pretty print a DeviceGraph with Devices followed by links."""
        lines = list()
        for device in self.devices:
            lines.append(str(device))
        for p0, p1 in self.links:
            lines.append(f"{p0} <--{self.links[frozenset({p0, p1})]}--> {p1}")
        return "\n".join(lines)

    def link(self, p0: DevicePort, p1: DevicePort,
             latency: str = '0s') -> None:
        """
        Link two DevicePorts with latency if provided.

        Links are bidirectional and the key is a frozenset of the two
        DevicePorts. Duplicate links (links between the same DevicePorts)
        are not permitted. Keep in mind that a unique DevicePort is created
        for each port number in a multi-port style port. If the link
        types to not match, then throw an exception. Devices that are linked
        to will be added to the graph automatically.
        Latency is expressed as a string with time units (ps, ns, us...)
        """
        if callable(p0) or callable(p1):
            raise RuntimeError(f"{p0} or {p1} is callable. This probably means"
                               f" you have a multi port and didn't pick a port"
                               f" number (ex. Device.portX(portNum))")

        def link_other_port(p0: DevicePort, p1: DevicePort) -> None:
            """Link a matching port through an expanding assembly."""
            if p0.link is not None:
                p2 = p0.link
                if not self.check_port_types(p1, p2):
                    raise RuntimeError(f'Port type mismatch {p1}, {p2}')
                # remove p0 from the links and connect p1 to p2
                p0.link = None
                p2.link = p1
                p1.link = p2
                self.ports.discard(p0)
                self.ports.add(p1)
                latency = self.links.pop(frozenset({p0, p2}))
                # add the other device to the graph
                self.add(p1.device)
                self.links[frozenset({p1, p2})] = latency

        if self.expanding is not None:
            if p0.device == self.expanding:
                link_other_port(p0, p1)
                return
            elif p1.device == self.expanding:
                link_other_port(p1, p0)
                return

        if p0 in self.ports or p1 in self.ports:
            raise RuntimeError(f'{p0} or {p1} already linked to')

        if not self.check_port_types(p0, p1):
            raise RuntimeError(f'Port type mismatch {p0}, {p1}')

        # Add devices to the graph
        self.add(p0.device)
        self.add(p1.device)

        # Storing the ports in a set so that we can quickly see if they
        # are linked to already
        self.ports |= {p0, p1}
        # Only update the links if neither are connected
        # otherwise we are most likely doing a separate graph expansion and
        # don't want to overwrite the port links that exist
        if p0.link is None and p1.link is None:
            p0.link = p1
            p1.link = p0
        self.links[frozenset({p0, p1})] = latency

    def add(self, device: Device, sub: bool = False) -> None:
        """
        Add a Device to the graph.

        The Device must be a ahp_graph Device. The name must be unique.
        If the Device has submodules, then we add those, as well.
        Do NOT add submodules to a Device after you have added it using
        this function, they will not be included in the DeviceGraph.
        """
        if device in self.devices:
            return

        if self.expanding is not None:
            device.name = f"{self.expanding.name}.{device.name}"
            if (self.expanding.partition is not None
                    and device.partition is None):
                device.partition = self.expanding.partition

        if device.name in self.names:
            raise RuntimeError(f'Device name {device.name} already in graph')

        self.devices.add(device)
        self.names.add(device.name)

        if device.subOwner is not None and not sub:
            dev = device
            while dev.subOwner is not None:
                dev = dev.subOwner
            self.add(dev)

        for (dev, _, _) in device.subs:
            self.add(dev, True)

    def count_devices(self) -> dict:
        """
        Count the Devices in a graph.

        Return a map of Devices to integer counts. The keys are of the
        form "CLASS_MODEL".
        """
        counter = collections.defaultdict(int)
        for device in self.devices:
            counter[device.get_category()] += 1
        return counter

    @staticmethod
    def check_port_types(p0: DevicePort, p1: DevicePort) -> bool:
        """Check that the port types for the two ports match."""
        t0 = p0.device.portinfo[p0.name][1]
        t1 = p1.device.portinfo[p1.name][1]
        return t0 == t1

    def verify_links(self) -> None:
        """Verify that all required ports are linked up."""
        # Create a map of Devices to all ports linked on those Devices.
        d2ports = collections.defaultdict(set)
        for p0, p1 in self.links:
            d2ports[p0.device].add(p0.name)
            d2ports[p1.device].add(p1.name)

        # Walk all Devices and make sure required ports are connected.
        for device in self.devices:
            for name, info in device.portinfo.items():
                if info[2] and name not in d2ports[device]:
                    raise RuntimeError(f"{device.name} requires port {name}")

    def check_partition(self) -> None:
        """Check to make sure the graph has ranks specified for all Devices."""
        for d in self.devices:
            if d.partition is None:
                raise RuntimeError(f"No partition for Device: {d.name}")

    def follow_links(self, rank: int, prune: bool = False) -> None:
        """
        Chase links between ranks.

        Follow links from the specified rank and expand assemblies until links
        are fully defined (links touch library Devices on both sides).
        Optional prune flag will remove unnecessary Devices and links from
        the graph. This will result in a different overall graph but will
        potentially save memory
        """
        self.check_partition()
        while True:
            devices = set()
            linksToRemove = list()
            for p0, p1 in self.links:
                # One of the Devices is on the rank that we are
                # following links on
                if (p0.device.partition[0] == rank
                        or p1.device.partition[0] == rank):
                    for p in [p0, p1]:
                        if p.device.library is None:
                            devices.add(p.device)
                # This link is not used on our rank, remove it
                # Devices will not be inserted into the graph if they
                # aren't linked to
                elif prune:
                    linksToRemove.append(frozenset({p0, p1}))

            # Can't remove items from the dictionary while we are iterating
            # through it, so we store them in a set and remove them after
            # should be better than copying the dictionary
            for p0, p1 in linksToRemove:
                del self.links[frozenset({p0, p1})]
                p0.link = None
                p1.link = None
                self.ports -= {p0, p1}

            if devices:
                self.flatten(expand=devices)
            else:
                break

    def flatten(self, levels: int = None, name: str = None,
                rank: int = None, expand: set = None) -> None:
        """
        Recursively flatten the graph by the specified number of levels.

        For example, if levels is one, then only one level of the hierarchy
        will be expanded. If levels is None, then the graph will be fully
        expanded.

        The name parameter lets you flatten the graph only under a specified
        Device. This expansion allows for multilevel expansion of an assembly
        of assemblies since Devices that are created during expansion have
        the parent assembly's name prepended to their own

        The rank parameter lets you flatten the graph for all Devices in the
        specified rank

        You can also provide a set of Devices to expand instead of looking
        through the entire graph
        """
        # Devices must have a matching name if provided, a matching
        # rank if provided, and be within the expand set if provided
        if levels == 0:
            return

        assemblies = set()
        if name is not None:
            splitName = name.split(".")

        # only check the expand set if provided
        if expand is not None:
            devs = expand
        else:
            devs = self.devices

        for dev in devs:
            assembly = dev.library is None
            if not assembly:
                continue

            # check to see if the name matches
            if name is not None:
                assembly &= splitName == dev.name.split(".")[0: len(splitName)]
            # rank to check
            if rank is not None:
                assembly &= rank == dev.partition[0]

            if assembly:
                assemblies.add(dev)

        if not assemblies:
            return

        # Expand the required Devices
        for device in assemblies:
            self.expanding = device
            device.expand(self)
            self.names.discard(device.name)
            self.devices.discard(device)
        self.expanding = None

        if expand is None:
            # Recursively flatten
            self.flatten(None if levels is None else levels-1, name, rank)

    def write_dot(self, name: str, draw: bool = False, ports: bool = False,
                  hierarchy: bool = True) -> None:
        """
        Take a DeviceGraph and write it as a graphviz dot graph.

        All output will be stored in a folder called output
        The draw parameter will automatically generate SVGs if set to True
        The ports parameter will display ports on the graph if set to True

        The hierarchy parameter specifies whether you would like to view the
        graph as a hierarchy of assemblies or if you would like get a flat
        view of the graph as it is.
        hierarchy is True by default, and highly recommended for large graphs
        """
        if not os.path.exists('output'):
            os.makedirs('output')

        if hierarchy:
            self.__write_dot_hierarchy(name, draw, ports)
        else:
            self.__write_dot_flat(name, draw, ports)

    def __write_dot_hierarchy(self, name: str, draw: bool = False,
                              ports: bool = False, assembly: str = None,
                              types: set = None) -> set:
        """
        Take a DeviceGraph and write dot files for each assembly.

        Write a graphviz dot file for each unique assembly (type, model) in the
        graph.
        assembly and types should NOT be specified by the user, they are
        soley used for recursion of this function
        """
        graph = self.__format_graph(name, ports)
        if types is None:
            types = set()

        splitName = None
        splitNameLen = None
        if assembly is not None:
            splitName = assembly.split('.')
            splitNameLen = len(splitName)

        # Expand all unique assembly types and write separate graphviz files
        for dev in self.devices:
            if dev.library is None:
                category = dev.get_category()
                if category not in types:
                    types.add(category)
                    expanded = DeviceGraph()
                    dev.expand(expanded)
                    types = expanded.__write_dot_hierarchy(category, draw,
                                                           ports, dev.name,
                                                           types)

        # Need to check if the provided assembly name is in the graph
        # and if so make that our cluster
        if assembly is not None:
            for dev in self.devices:
                if assembly == dev.name:
                    # This device is the assembly that we just expanded
                    # make this a cluster and add its ports as nodes
                    clusterName = f"cluster_{dev.type}"
                    subgraph = graph.subgraph(name=clusterName, color='green')
                    for port in dev.ports:
                        graph.add_node(f"{dev.type}:{port}",
                                       shape='diamond', label=port,
                                       color='green', fontcolor='green')
                    break
        else:
            # No provided assembly, this is most likely the top level
            subgraph = graph

        # Loop through all Devices and add them to the graphviz graph
        for dev in self.devices:
            if assembly != dev.name:
                label = dev.name
                nodeName = dev.name
                if assembly is not None:
                    if splitName == dev.name.split('.')[0:splitNameLen]:
                        nodeName = '.'.join(dev.name.split('.')[splitNameLen:])
                        label = nodeName
                if dev.model is not None:
                    label += f"\\nmodel={dev.model}"
                if ports:
                    portLabels = dev.label_ports()
                    if portLabels != '':
                        label += f"|{portLabels}"

                # If the Device is an assembly, put a link to its SVG
                if dev.library is None:
                    subgraph.add_node(nodeName, label=label,
                                      href=f"{dev.get_category()}.svg",
                                      color='blue', fontcolor='blue')
                elif dev.subOwner is not None:
                    subgraph.add_node(nodeName, label=label,
                                      color='purple', fontcolor='purple')
                else:
                    subgraph.add_node(nodeName, label=label)

        self.__dot_add_links(graph, ports, assembly, splitName, splitNameLen)

        graph.write(f"output/{name}.dot")
        if draw:
            graph.draw(f"output/{name}.svg", format='svg', prog='dot')

        return types

    def __write_dot_flat(self, name: str, draw: bool = False,
                         ports: bool = False) -> None:
        """
        Write the DeviceGraph as a DOT file.

        It is suggested that you use write_dot_hierarchy for large graphs
        """
        graph = self.__format_graph(name, ports)

        for dev in self.devices:
            label = dev.name
            if dev.model is not None:
                label += f"\\nmodel={dev.model}"
            if ports:
                portLabels = dev.label_ports()
                if portLabels != '':
                    label += f"|{portLabels}"
            if dev.subOwner is not None:
                graph.add_node(dev.name, label=label,
                               color='purple', fontcolor='purple')
            else:
                graph.add_node(dev.name, label=label)

        self.__dot_add_links(graph, ports)

        graph.write(f"output/{name}.dot")
        if draw:
            graph.draw(f"output/{name}.svg", format='svg', prog='dot')

    @staticmethod
    def __format_graph(name: str, record: bool = False) -> pygraphviz.AGraph:
        """Format a new graph."""
        h = ('.edge:hover text {\n'
             '\tfill: red;\n'
             '}\n'
             '.edge:hover path, .node:hover polygon, .node:hover ellipse {\n'
             '\tstroke: red;\n'
             '\tstroke-width: 10;\n'
             '}')
        if not os.path.exists('output/highlightStyle.css'):
            with open('output/highlightStyle.css', 'w') as f:
                f.write(h)

        graph = pygraphviz.AGraph(strict=False, name=name)
        graph.graph_attr['stylesheet'] = 'highlightStyle.css'
        graph.node_attr['style'] = 'filled'
        graph.node_attr['fillcolor'] = '#EEEEEE'  # light gray fill
        graph.edge_attr['penwidth'] = '2'
        if record:
            graph.node_attr['shape'] = 'record'
            graph.graph_attr['rankdir'] = 'LR'

        return graph

    def __dot_add_links(self, graph, ports: bool = False,
                        assembly: str = None, splitName: list = None,
                        splitNameLen: int = None) -> None:
        """Add edges to the graph with a label for the number of edges."""
        def port2Node(port: DevicePort) -> str:
            """Return a node name given a DevicePort."""
            node = port.device.name
            if node == assembly:
                return f"{port.device.type}:{port.name}"
            elif assembly is not None:
                if splitName == node.split('.')[0:splitNameLen]:
                    node = '.'.join(node.split('.')[splitNameLen:])
            if ports:
                return (node, port.name)
            else:
                return node

        # Create a list of all of the links
        links = list()
        for p0, p1 in self.links:
            links.append((port2Node(p0), port2Node(p1)))

        # Setup a counter so we can check for duplicates
        duplicates = collections.Counter(links)
        for key in duplicates:
            label = ''
            if duplicates[key] > 1:
                label = str(duplicates[key])

            graphNodes = [key[0], key[1]]
            graphPorts = ['', '']
            for i in range(2):
                if type(key[i]) is tuple:
                    graphNodes[i] = key[i][0]
                    graphPorts[i] = key[i][1]
            # Add edges using the number of links as a label
            graph.add_edge(graphNodes[0], graphNodes[1], label=label,
                           tailport=graphPorts[0], headport=graphPorts[1])

        def device2Node(dev: Device) -> str:
            """Return a node name given a Device."""
            node = dev.name
            if assembly is not None:
                if splitName == node.split('.')[0:splitNameLen]:
                    node = '.'.join(node.split('.')[splitNameLen:])
            return node

        # Add "links" to submodules so they don't just float around
        for dev in self.devices:
            if dev.subOwner is not None:
                graph.add_edge(device2Node(dev), device2Node(dev.subOwner),
                               color='purple', style='dashed')



"""
Server assembly.

Constructed from Processors and main memory using a NoC.
"""
from ahp_graph.Device import *
from ahp_graph.DeviceGraph import *
from processor import *


class MemNIC(Device):
    """MemNIC."""

    library = 'memHierarchy.MemNIC'
    portinfo = PortInfo()
    portinfo.add('port', 'simpleNet')
    attr = {
        "network_bw": "25GB/s"
    }

    def __init__(self, name: str, model: str, attr: dict = None) -> None:
        """Initialize with a model describing where this is on the NoC."""
        if model == 'Cache':
            parameters = {
                "group": 1,
                "destinations": "2,3"  # DirCtrl, SHMEMNIC
            }
        elif model == 'DirCtrl':
            parameters = {
                "group": 2,
                "sources": "1,3"  # Cache, SHMEMNIC
            }
        elif model == 'SHMEMNIC':
            parameters = {
                "group": 3,
                "sources": "1",  # Cache
                "destinations": "2"  # DirCtrl
            }
        else:
            return None

        if attr is not None:
            parameters.update(attr)
        super().__init__(name, attr=parameters)


class Router(Device):
    """Router."""

    library = 'merlin.hr_router'
    portinfo = PortInfo()
    portinfo.add('port', 'simpleNet', None, False, '#')

    def __init__(self, name: str, model: str, id: int = 0, ports: int = 1,
                 attr: dict = None) -> None:
        """Initialize with a model for a NoC or Interconnect."""
        parameters = {
            "id": id,
            "num_ports": ports
        }
        if model == 'NoC':
            parameters.update({
                "flit_size": "72B",
                "xbar_bw": "50GB/s",
                "link_bw": "25GB/s",
                "input_buf_size": "40KB",
                "output_buf_size": "40KB"
            })
        elif model == 'Interconnect':
            parameters.update({
                "link_bw": "16GB/s",
                "xbar_bw": "16GB/s",
                "link_lat": "10ns",
                "input_latency": "10ns",
                "output_latency": "10ns",
                "flit_size": "8B",
                "input_buf_size": "14KB",
                "output_buf_size": "14KB",
                "num_dims": 2
            })
        else:
            return None

        if attr is not None:
            parameters.update(attr)
        super().__init__(name, model, parameters)


class SingleRouter(Device):
    """Single Router Topology."""

    library = 'merlin.singlerouter'


class DirectoryController(Device):
    """DirectoryController."""

    library = 'memHierarchy.DirectoryController'
    portinfo = PortInfo()
    portinfo.add('direct_link', 'simpleMem', required=False)
    attr = {
        "coherence_protocol": "MESI",
        "entry_cache_size": 1024,
        "addr_range_start": 0x0,
        "addr_range_end": 0x7fffffff
    }


class MemController(Device):
    """MemController."""

    library = 'memHierarchy.MemController'
    portinfo = PortInfo()
    portinfo.add('direct_link', 'simpleMem', required=False)
    attr = {
        "clock": "1GHz",
        "backend.mem_size": "4GiB",
        "backing": "malloc",
        "addr_range_start": 0x0,
        "addr_range_end": 0x7fffffff
    }


class simpleMem(Device):
    """simpleMem."""

    library = 'memHierarchy.simpleMem'
    attr = {
        "mem_size": "2GiB",
        "access_time": "1 ns"
    }


class RDMA_NIC(Device):
    """RDMA_NIC."""

    library = 'rdmaNic.nic'
    portinfo = PortInfo()
    portinfo.add('port', 'simpleNet', None, False, '#')
    attr = {
        "clock": "8GHz",
        "maxPendingCmds": 128,
        "maxMemReqs": 256,
        "maxCmdQSize": 128,
        "cache_line_size": 64,
        'baseAddr': 0x80000000,
        'cmdQSize': 64
    }

    def __init__(self, name: str, id: int = 0, nodes: int = 2,
                 cores: int = 1, attr: dict = None) -> None:
        """
        Initialize using various simulation parameters.

        Need to know the number of processing elements (cores) per node.
        Also need the total number of nodes in the system.
        """
        super().__init__(name, attr=attr)
        self.attr['pesPerNode'] = cores
        self.attr['nicId'] = id
        self.attr['numNodes'] = nodes


class LinkControl(Device):
    """LinkControl."""

    library = 'merlin.linkcontrol'
    portinfo = PortInfo()
    portinfo.add('rtr_port', 'simpleNet')
    attr = {
        "link_bw": "16GB/s",
        "input_buf_size": "14KB",
        "output_buf_size": "14KB"
    }


class Server(Device):
    """Server constructed of a processor and some memory."""

    portinfo = PortInfo()
    portinfo.add('network', 'simpleNet')

    def __init__(self, name: str, node: int = 0,
                 racks: int = 2, nodes: int = 2, cores: int = 1,
                 attr: dict = None) -> None:
        """Store our name and use number of cores as the model type."""
        super().__init__(name, f"{cores}Core", attr)
        self.attr['node'] = node
        self.attr['racks'] = racks
        self.attr['nodes'] = nodes
        self.attr['cores'] = cores

    def expand(self, graph: DeviceGraph) -> None:
        """Expand the server into its components."""
        # Setup the NoC first so we can connect the Processors to it
        # Device names created by an assembly will automatically have the
        # assembly name prefixed to the name provided.
        NoC = Router("NoC", 'NoC', 0, self.attr['cores'] + 2)
        NoC_topo = SingleRouter("NoC_topo")
        NoC.add_submodule(NoC_topo, 'topology')

        # Generate the appropriate number of Processors and L2 Caches
        for core in range(self.attr['cores']):
            cpu = Processor(f"CPU{core}", core, self.attr['cores'])

            L2 = Cache(f"CPU{core}_L2", 'L2')
            L1_to_L2 = MemLink(f"CPU{core}_L1_to_L2")
            L2_to_mem = MemNIC(f"CPU{core}_L1_to_mem", 'Cache')
            L2.add_submodule(L1_to_L2, 'cpulink')
            L2.add_submodule(L2_to_mem, 'memlink')

            graph.link(cpu.low_network(0), L1_to_L2.port('port'), '1ns')
            graph.link(L2_to_mem.port('port'), NoC.port('port', core), '1ns')

        # Setup the main memory with controllers
        dirctrl = DirectoryController("DirectoryController")
        dir_to_mem = MemLink("dir_to_mem")
        dirNIC = MemNIC("dirNIC", 'DirCtrl')
        dirctrl.add_submodule(dir_to_mem, 'memlink')
        dirctrl.add_submodule(dirNIC, 'cpulink')

        memctrl = MemController("MemController")
        mem_to_dir = MemLink("mem_to_dir")
        memory = simpleMem("simpleMem")
        memctrl.add_submodule(mem_to_dir, 'cpulink')
        memctrl.add_submodule(memory, 'backend')

        # Initialize the RDMA_NIC and its interfaces
        nic = RDMA_NIC("NIC", self.attr['node'],
                       self.attr['racks'] * self.attr['nodes'],
                       self.attr['cores'])
        mmioIf = memInterface("MMIO_IF")
        mmioNIC = MemNIC("MMIO_NIC", 'SHMEMNIC')
        netLink = LinkControl("netLink")
        mmioIf.add_submodule(mmioNIC, 'memlink')
        nic.add_submodule(mmioIf, 'mmio')
        nic.add_submodule(netLink, 'rtrLink')

        graph.link(NoC.port('port', None), dirNIC.port('port'), '1ns')
        graph.link(NoC.port('port', None), mmioNIC.port('port'), '10ns')
        graph.link(dir_to_mem.port('port'), mem_to_dir.port('port'), '1ns')

        # Our external link is through the RDMA_NIC
        # Generally you don't want to put latency on the links to assembly
        # ports (ex: self.port) and allow whatever uses the assembly to
        # specify latency for the connection (it will get ignored anyway)
        graph.link(netLink.port('rtr_port'), self.network)



"""Simple example of two SST components playing pingpong with messages."""

from ahp_graph.DeviceGraph import *
from ahp_graph.SSTGraph import *
from architecture import architecture


if __name__ == "__main__":
    """
    If we are running as a script (either via Python or from SST), then
    proceed.  Check if we are running with SST or from Python.
    """
    import argparse
    try:
        import sst
        SST = True
    except ImportError:
        SST = False

    parser = argparse.ArgumentParser(description='PingPong')
    parser.add_argument('--num', type=int, default=2,
                        help='how many pingpongs to include')
    parser.add_argument('--repeats', type=int, default=5,
                        help='how many message volleys to run')
    parser.add_argument('--partitioner', type=str, default='sst',
                        help='which partitioner to use: ahp_graph, sst')
    args = parser.parse_args()

    # Construct a DeviceGraph with the specified architecture
    graph = architecture(args.repeats, args.num)
    sstgraph = SSTGraph(graph)

    if SST:
        # If running within SST, generate the SST graph
        # There are multiple ways to run, below are two examples

        # SST partitioner
        # This will work in serial or running SST with MPI in parallel
        if args.partitioner.lower() == 'sst':
            sstgraph.build()

        # MPI mode with ahp_graph graph partitioning. Specifying nranks tells
        # ahp_graph that it is doing the partitioning, not SST
        # For this to work you need to pass --parallel-load=SINGLE to sst
        elif args.partitioner.lower() == 'ahp_graph':
            sstgraph.build(args.num)

    else:
        # generate a graphviz dot file including the hierarchy
        graph.write_dot('pingpong', draw=True, ports=True)
        sstgraph.write_json('pingpong.json', nranks=args.num)

        # generate a flat graphviz dot file and json output for demonstration
        graph.flatten()
        graph.write_dot('pingpongFlat', draw=True, ports=True, hierarchy=False)
        sstgraph.write_json('pingpongFlat.json')



"""Collection of Unit tests for ahp_graph Device."""

from ahp_graph.Device import *
from Devices import *


def test_library() -> None:
    """Test of Device with Library."""
    ltd = LibraryTestDevice().library
    assert ltd == 'ElementLibrary.Component', ltd


def test_port() -> None:
    """Test of Device with various ports."""
    ptd = PortTestDevice()

    assert ptd.portinfo['default'][0] == 1, 'default limit'
    assert ptd.portinfo['default'][1] is None, 'default type'
    assert ptd.portinfo['default'][2] is True, 'default required'
    assert ptd.portinfo['default'][3] == '.#', 'default format'

    assert ptd.portinfo['no_limit'][0] is None, 'No limit'
    assert ptd.portinfo['limit'][0] == 2, 'limit'
    assert ptd.portinfo['ptype'][1] == 'test', 'type'
    assert ptd.portinfo['optional'][2] is False, 'optional'
    assert ptd.portinfo['format'][3] == '(#)', 'format'

    assert ptd.default == ptd.port('default'), '__getattr__'
    assert ptd.default.device == ptd, 'device'  # type: ignore[union-attr]
    assert ptd.default.name == 'default', 'name'  # type: ignore[union-attr]
    assert ptd.default.number is None, 'No port number'  # type: ignore[union-attr]
    notDefinedTest = None
    try:
        ptd.portNotDefined
        notDefinedTest = False
    except RuntimeError:
        notDefinedTest = True
    assert notDefinedTest, 'port not defined'

    noLimitTest = None
    try:
        for i in range(1000):
            ptd.no_limit(None)  # type: ignore[operator]
        ptd.no_limit(3004823048)  # type: ignore[operator]
        noLimitTest = True
    except RuntimeError:
        noLimitTest = False
    assert noLimitTest, 'no limit instantiation'

    limitTest = None
    try:
        ptd.limit(2)  # type: ignore[operator]
        limitTest = False
    except RuntimeError:
        limitTest = True
    assert limitTest, 'limit instantiation'

    assert ptd.format(0).get_name() == 'format(0)', 'format incorrect'  # type: ignore[operator]
    assert ptd.port('port').name == 'port', 'port named port'


def test_assembly() -> None:
    """Test of Device that is an assembly."""
    ratd = RecursiveAssemblyTestDevice(0).library is None
    assert ratd, 'assembly'

    results = None
    try:
        class AssemblyNoExpandTestDevice(Device):
            """Assembly that doesn't define the expand function."""

            def __init__(self) -> None:
                """Test Device for assembly with no expand function."""
                super().__init__('AssemblyNoExpandTestDevice')

        anetd = AssemblyNoExpandTestDevice()
        results = False
    except RuntimeError:
        results = True
    assert results, 'no expand method'


def test_model() -> None:
    """Test of device with model."""
    mtd = ModelTestDevice('model0')
    ltd = LibraryTestDevice()
    assert mtd.model == 'model0', 'model'
    assert ltd.model is None, 'None model'


def test_attribute() -> None:
    """Test of Device with attributes."""
    attr1 = {'a1': 1, 'a2': 'blue', 'a3': False}
    atd1 = AttributeTestDevice({})
    assert atd1.attr == attr1, 'class attr'
    attr2 = {'a1': 1, 'a2': 'blue', 'a3': False, 'b': 'test'}
    atd2 = AttributeTestDevice({'b': 'test'})
    assert atd2.attr == attr2, 'instance attr'
    assert atd1.attr == attr1, 'class attr unchanged'


def test_partition() -> None:
    """Test of partitioning a device."""
    ratd = RecursiveAssemblyTestDevice(0)
    ratd.set_partition(1)
    assert ratd.partition[0] == 1, 'rank'  # type: ignore[index]
    assert ratd.partition[1] is None, 'thread None'  # type: ignore[index]
    ratd.set_partition(2, 3)
    assert ratd.partition[0] == 2, 'rank'  # type: ignore[index]
    assert ratd.partition[1] == 3, 'thread'  # type: ignore[index]


def test_category() -> None:
    """Test of device category output."""
    ltd = LibraryTestDevice()
    mtd = ModelTestDevice('model')
    assert ltd.get_category() == ltd.name, 'No model'
    assert mtd.get_category() == f'{mtd.name}_{mtd.model}', 'model'


def test_submodule() -> None:
    """Test for adding a submodule to a device."""
    ltd0 = LibraryTestDevice()
    ltd1 = LibraryTestDevice()
    ltd0.add_submodule(ltd1, 'slotName')
    pop = ltd0.subs.pop()
    assert ltd1.subOwner == ltd0, 'subOwner'
    assert pop[0] == ltd1, 'subs'
    assert pop[1] == 'slotName', 'slotName'
    assert pop[2] is None, 'slotIndex'



"""
Abstract HPC cluster.

This demonstrates how assemblies in ahp_graph can be used to build systems
out of small components.
This example utilizes sst-elements to start with a processor and memory
and then build all the way up to a cluster.
"""
from ahp_graph.Device import *
from ahp_graph.DeviceGraph import *
from ahp_graph.SSTGraph import *
from server import *


class TorusTopology(Device):
    """Torus Topology."""

    library = 'merlin.torus'
    attr = {
        "width": "1x1"
    }

    def __init__(self, name: str, shape: str = '2x2', nodes: int = 1,
                 attr: dict = None) -> None:
        """Initialize using the shape as the model."""
        super().__init__(name, shape, attr)
        self.attr['shape'] = shape
        self.attr['local_ports'] = nodes


class Rack(Device):
    """Rack constructed of a router and some servers."""

    portinfo = PortInfo()
    portinfo.add('network', 'simpleNet', None, False)

    def __init__(self, name: str, shape: str = '2x2', rack: int = 0,
                 nodes: int = 1, cores: int = 1,
                 attr: dict = None) -> None:
        """Store our name and the model describing nodes and cores per node."""
        super().__init__(name, f"{nodes}Node_{cores}Core", attr)
        self.attr['shape'] = shape
        self.attr['rack'] = rack
        dimX = int(shape.split('x')[0])
        dimY = int(shape.split('x')[1])
        self.attr['racks'] = dimX * dimY
        self.attr['nodes'] = nodes
        self.attr['cores'] = cores

    def expand(self, graph: DeviceGraph) -> None:
        """Expand the rack into its components."""
        # Device names created by an assembly will automatically have the
        # assembly name prefixed to the name provided.
        router = Router("Router", 'Interconnect',
                        self.attr['rack'], self.attr['nodes'] + 4)
        topology = TorusTopology("TorusTopology",
                                 self.attr['shape'], self.attr['nodes'])
        router.add_submodule(topology, "topology")
        router.set_partition(self.partition[0], 0)

        # make all the torus ports available outside the rack
        for i in range(4):
            # Generally you don't want to put latency on the links to assembly
            # ports (ex: self.port) and allow whatever uses the assembly to
            # specify latency for the connection (it will get ignored anyway)
            graph.link(router.port('port', i), self.network(i))

        # connect the servers to the router
        for node in range(self.attr['nodes']):
            server = Server(f"Server{node}",
                            (self.attr['rack'] * self.attr['nodes']) + node,
                            self.attr['racks'], self.attr['nodes'],
                            self.attr['cores'])
            server.set_partition(self.partition[0], node+1)
            graph.link(router.port('port', None), server.network, '10ns')


def Cluster(shape: str = '2x2', nodes: int = 1,
            cores: int = 1) -> DeviceGraph:
    """HPC Cluster built out of racks. Using a 2D torus network."""
    graph = DeviceGraph()  # initialize a Device Graph
    dimX = int(shape.split('x')[0])
    dimY = int(shape.split('x')[1])

    racks = dict()
    for x in range(dimX):
        for y in range(dimY):
            racks[f"{x}x{y}"] = Rack(f"Rack{x}x{y}", shape, (x * dimY) + y,
                                     nodes, cores)
            racks[f"{x}x{y}"].set_partition((x * dimY) + y)

    # initialize the four torus ports
    # order is 0: north, 1: east, 2: south, 3: west
    for x in range(dimX):
        for y in range(dimY):
            graph.link(racks[f"{x}x{y}"].network(1),
                       racks[f"{(x+1) % dimX}x{y}"].network(3), '10ns')

            graph.link(racks[f"{x}x{y}"].network(0),
                       racks[f"{x}x{(y+1) % dimY}"].network(2), '10ns')

    return graph


if __name__ == "__main__":
    """
    If we are running as a script (either via Python or from SST), then
    proceed. Check if we are running with SST or from Python.
    """
    import argparse
    try:
        import sst
        SST = True
    except ImportError:
        SST = False

    parser = argparse.ArgumentParser(
        description='HPC Cluster Simulation')
    parser.add_argument('--shape', type=str, default='2x2',
                        help='optional shape to use for the topology')
    parser.add_argument('--nodes', type=int, default=1,
                        help='optional number of nodes per rack')
    parser.add_argument('--cores', type=int, default=1,
                        help='optional number of cores per server')
    parser.add_argument('--partitioner', type=str, default='sst',
                        help='which partitioner to use: ahp_graph, sst')
    args = parser.parse_args()

    dims = [int(x) for x in args.shape.split('x')]
    racks = dims[0] * dims[1]

    # Create a cluster with the given parameters
    graph = Cluster(args.shape, args.nodes, args.cores)
    sstgraph = SSTGraph(graph)

    if SST:
        # If running within SST, generate the SST graph
        # There are multiple ways to run, below are a few common examples

        # SST partitioner
        # This will work in serial or running SST with MPI in parallel
        if args.partitioner.lower() == 'sst':
            sstgraph.build()

        # MPI mode with ahp_graph graph partitioning. Specifying nranks tells
        # BuildSST that it is doing the partitioning, not SST
        # For this to work you need to pass --parallel-load=SINGLE to sst
        elif args.partitioner.lower() == 'ahp_graph':
            sstgraph.build(racks)

    else:
        # generate a graphviz dot file and json output for demonstration
        graph.write_dot('cluster', draw=True, ports=True)
        sstgraph.write_json('cluster.json', racks)
